unit TypedEdit;

{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/NPL/NPL-1_1Final.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: TypedEdit.pas, released November 1999.

The Initial Developer of the Original Code is Anthony Steele.
Portions created by Anthony Steele are Copyright (C) 1999 Anthony Steele.
All Rights Reserved.

Modified:
  Version 0.4 April 2000 to use jcl code library where possible
  2 Jun 2K : used StripNonNumberChars in TCustomCurrencyEdit.GetValue

Last Modified: 02 Jun 2000
Current Version: 0.4


Comments - controls to capure numbers of various kinds (Int, Float, Money)
 The money edit also does some display formatting

   Still to do

    - test some more

   - The controls do not work well if zero is not within the min/max range
     What should happen when 0 is not in the range?
     control should start up with ? min ?max ? 0? displayed
     User should set the value?

     Consider only checking the full range on exit.
     During data capture, extend the range to include 0

     - See if a base class is useful for these edit controls
     - Display formatting for int & float edits?

     The code in these controls is a bit repetitive, but unfortunately
     you can't have a base class property with polymorphic type

}


interface

uses
  { delphi } Classes,
  { local } MemoEx;

type

  TCustomIntegerEdit = class(TCustomAutoSelectMemo)
  private
    { property implementation }
    fiMaxValue, fiMinValue: integer;
    fbHasMaxValue, fbHasMinValue: boolean;

  protected
    { overrides }
    procedure KeyPress(var key: char); override;

    { property implementation }
    procedure SetValue(iNewValue: integer); virtual;
    function GetValue: integer; virtual;

    procedure SetMinValue(iNewValue: integer);
    procedure SetMaxValue(iNewValue: integer);

   { worker procs }
    function ChopToRange(iValue: integer): integer;
    function IsInRange(iValue: integer): boolean; overload;
    function IsInRange(sValue: string): boolean; overload;

    property MaxValue: integer read fiMaxValue write fiMaxValue;
    property MinValue: integer read fiMinValue write fiMinValue;

    property HasMaxValue: boolean read fbHasMaxValue write fbHasMaxValue;
    property HasMinValue: boolean read fbHasMinValue write fbHasMinValue;

  public
    constructor Create(AOwner: TComponent); override;
    procedure Assign(Source: TPersistent); override;

    property Value: integer read GetValue write SetValue;

  end;

  { to capture and display float values }
  TCustomFloatEdit = class(TCustomAutoSelectMemo)
  private
    { property implementation }
    feMaxValue, feMinValue: extended;
    fbHasMaxValue, fbHasMinValue: boolean;
    fiMaxDecimals: integer;

  protected
    { overrides }
    procedure KeyPress(var key: char); override;

    { property implementation }
    procedure SetValue(eNewValue: extended);
    function GetValue: extended;

    procedure SetMinValue(eNewValue: extended);
    procedure SetMaxValue(eNewValue: extended);

    procedure SetMaxDecimals(iNewValue: integer);

   { worker procs }
    function ChopToRange(eValue: extended): extended;
    function IsInRange(eValue: extended): boolean; overload;
    function IsInRange(sValue: string): boolean; overload;

    property MaxValue: extended read feMaxValue write feMaxValue;
    property MinValue: extended read feMinValue write feMinValue;
    property MaxDecimals: integer read fiMaxDecimals write SetMaxDecimals;

    property HasMaxValue: boolean read fbHasMaxValue write fbHasMaxValue;
    property HasMinValue: boolean read fbHasMinValue write fbHasMinValue;

  public
    constructor Create(AOwner: TComponent); override;
    procedure Assign(Source: TPersistent); override;

    property Value: extended read GetValue write SetValue;
  end;


  { control to capture a year
    int edit with automatic 2 digit -> 4 digit expansion
    in a range controlld by the PivotYear
    which is initialised from the program/system settings  }
  TCustomYearEdit = class(TCustomIntegerEdit)
  private
    fiPivotYear: integer;
  protected
    procedure DoExit; override;

    procedure SetValue(iNewValue: integer); override;
    function GetValue: integer; override;

    procedure FixYear;

  public
    constructor Create(AOwner: TComponent); override;
    procedure Assign(Source: TPersistent); override;

    property PivotYear: integer read fiPivotYear write fiPivotYear;

  end;

{ to capture and display currency values
    when the control has focus, the number is displayed simply
    when the control does not have focus, the number is formatted
    as per control panel currency settings
  }
  TCustomCurrencyEdit = class(TCustomAutoSelectMemo)
  private
    { property implementation }
    feMaxValue, feMinValue: currency;
    fbHasMaxValue, fbHasMinValue: boolean;

  protected
    { overrides }
    procedure KeyPress(var key: char); override;

    { property implementation }
    procedure SetValue(eNewValue: currency); virtual;
    function GetValue: currency;

    procedure SetMinValue(eNewValue: currency);
    procedure SetMaxValue(eNewValue: currency);


   { worker procs }
    function ChopToRange(eValue: currency): currency;
    function IsInRange(eValue: currency): boolean; overload;
    function IsInRange(sValue: string): boolean; overload;

    property MaxValue: currency read feMaxValue write feMaxValue;
    property MinValue: currency read feMinValue write feMinValue;

    property HasMaxValue: boolean read fbHasMaxValue write fbHasMaxValue;
    property HasMinValue: boolean read fbHasMinValue write fbHasMinValue;

  public
    constructor Create(AOwner: TComponent); override;
    procedure Assign(Source: TPersistent); override;

    property Value: currency read GetValue write SetValue;
  end;


 { another bit of functionality
    - when the control does not have focus, display a formatted currency string
     (as per control panel)
    - when it has focus, display a raw number string for editing

    To do
    - Extend this functionality to int & float edits?

    I would put code this in as a base class below all the number edits,
    but unfortunately it uses the Value property which
    is different per subclass.
    it could easily be duplicated for int & float edit controls
    but I don't want to go there until I am convinced that there is no other way

    - Make the format mask a published property, not a constant?
      this is more applicable to Int & Float edits
    }
  TCustomFormattedCurrencyEdit = class(TCustomCurrencyEdit)
  private

  protected
    procedure FormatText; virtual;
    procedure UnformatText; virtual;

    procedure SetValue(eNewValue: currency); override;

    procedure DoEnter; override;
    procedure DoExit; override;

  end;

{ as required by Jedi VCL standards }
const
  TINTEGEREDIT_VERSION   = 0.4;
  TFLOATEDIT_VERSION     = 0.4;
  TYEAR_EDIT_VERSION     = 0.4;
  TCURRENCY_EDIT_VERSION = 0.4;

type

  TIntegerEdit = class(TCustomIntegerEdit)
  published
    { most properties of TMemo
     no Text or WordWrap properties }
    property Align;
    property Alignment;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property WantReturns;
    property WantTabs;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;

    { new properties }
    property AutoSelect;
    property Value;
    property MaxValue;
    property MinValue;
    property HasMaxValue;
    property HasMinValue;

  end;


  TFloatEdit = class(TCustomFloatEdit)
  published
    { most properties of TMemo
     no Text or WordWrap properties }
    property Align;
    property Alignment;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property WantReturns;
    property WantTabs;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;

    { new properties }
    property AutoSelect;
    property Value;
    property MaxValue;
    property MinValue;
    property MaxDecimals;

    property HasMaxValue;
    property HasMinValue;

  end;

  TYearEdit = class(TCustomYearEdit)
  published
    { most properties of TMemo
     no Text or WordWrap properties }
    property Align;
    property Alignment;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property WantReturns;
    property WantTabs;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;

    { new properties }
    property AutoSelect;
    property Value;
    property MaxValue;
    property MinValue;
    property HasMaxValue;
    property HasMinValue;
    property PivotYear;

  end;

  { surface class }
  TCurrencyEdit = class(TCustomFormattedCurrencyEdit)

  published
    { most properties of TMemo
     no Text or WordWrap properties }
    property Align;
    property Alignment;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property WantReturns;
    property WantTabs;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;

    { new properties }
    property AutoSelect;
    property Value;
    property MaxValue;
    property MinValue;

    property HasMaxValue;
    property HasMinValue;
  end;

implementation

uses
  { delphi } SysUtils,
  { jcl } JCLStrings, jclDatetime,
  { local } VCLFunctions;

{--------------------------------------------------------------------------------
  Custom Integer Edit }

function TCustomIntegerEdit.ChopToRange(iValue: integer): integer;
begin
  Result := iValue;

  if HasMaxValue and (iValue > MaxValue) then
    Result := fiMaxValue;

  if HasMinValue and (iValue < MinValue) then
    Result := fiMinValue;
end;

constructor TCustomIntegerEdit.Create(AOwner: TComponent);
begin
  inherited;

  { default is right aligned }
  Alignment := taRightJustify;
  Text      := '';

  { by default no min & max }
  fiMinValue    := 0;
  fiMaxValue    := 0;
  fbHasMinValue := False;
  fbHasMaxValue := False;
end;

procedure TCustomIntegerEdit.Assign(Source: TPersistent);
var
  lcSource: TCustomIntegerEdit;
begin
  inherited;

  if Source is TCustomIntegerEdit then
  begin
    lcSource    := Source as TCustomIntegerEdit;
    MinValue    := lcSource.MinValue;
    MaxValue    := lcSource.MaxValue;
    HasMinValue := lcSource.HasMinValue;
    HasMaxValue := lcSource.HasMaxValue;
  end;
end;

function TCustomIntegerEdit.GetValue: integer;
begin
  if Text = '' then
    Result := 0
  else
    Result := StrToIntDef(Text, 0);
end;

function TCustomIntegerEdit.IsInRange(iValue: integer): boolean;
begin
  Result := True;

  if HasMaxValue and (iValue > MaxValue) then
    Result := False;

  if HasMinValue and (iValue < MinValue) then
    Result := False;
end;

function TCustomIntegerEdit.IsInRange(sValue: string): boolean;
var
  liValue: integer;
begin
  if sValue = '' then
    liValue := 0
  else
    liValue := StrToIntDef(sValue, 0);

  Result := IsInRange(liValue);
end;

procedure TCustomIntegerEdit.KeyPress(var key: char);
var
  lsNewText: string;
begin
  { not intersted in control chars }
  if (Ord(Key)) < Ord(' ') then
  begin
    inherited KeyPress(Key);
    exit;
  end;

  { allow only digits and minus sign }
  if not CharIsDigit(Key) and (Key <> '-') then
    Key := #0;

  if Key = '-' then
  begin
    { are any neg. numbers in range ? }
    if HasMinValue and (MinValue >= 0) then
      Key := #0
    { minus sign only as first char }
    else if SelStart <> 0 then
      Key := #0;

    { only 1 minus sign }
    if StrLeft(lsNewText, 2) = '--' then
      Key := #0;
  end;

  { no leading zeros }
  if (SelStart = 0) and (Key = '0') and (StrLeft(lsNewText, 1) = '0') then
    Key := #0;

  { no leading '-0' }
  if (SelStart = 1) and (Key = '0') and (StrLeft(lsNewText, 2) = '-0') then
    Key := #0;

  { disallow the keypress if the value would go out of range }
  lsNewText := GetChangedText(Text, SelStart, SelLength, Key);
  if not IsInRange(lsNewText) then
    Key := #0;

  if Key <> #0 then
    inherited;
end;

procedure TCustomIntegerEdit.SetMaxValue(iNewValue: integer);
begin
  fiMaxValue := iNewValue;

  { make min value consistent }
  if MinValue > MaxValue then
    fiMinValue := MaxValue;
end;

procedure TCustomIntegerEdit.SetMinValue(iNewValue: integer);
begin
  fiMinValue := iNewValue;

  { make max value consistent }
  if MaxValue < MaxValue then
    fiMaxValue := MinValue;
end;

procedure TCustomIntegerEdit.SetValue(iNewValue: integer);
begin
  Text := IntToStr(ChopToRange(iNewValue));
end;

{-------------------------------------------------------------------------------
   TCustomFloatEdit }

const
{ The extended data type handles up to 20 digits - no point in
   setting max decimals to more than this }
  FLOAT_EDIT_MAX_MAX_DECIMALS = 20;

{ most data capture is more like this -
 typing in 19 decimal digits usually indicates forehead-on-the-keyboard syndrome }
  FLOAT_EDIT_DEFAULT_MAX_DECIMALS = 5;



procedure TCustomFloatEdit.Assign(Source: TPersistent);
var
  lcSource: TCustomFloatEdit;
begin
  inherited;

  if Source is TCustomFloatEdit then
  begin
    lcSource    := Source as TCustomFloatEdit;
    MinValue    := lcSource.MinValue;
    MaxValue    := lcSource.MaxValue;
    MaxDecimals := lcSource.MaxDecimals;

    HasMinValue := lcSource.HasMinValue;
    HasMaxValue := lcSource.HasMaxValue;
  end;
end;

function TCustomFloatEdit.ChopToRange(eValue: extended): extended;
begin
  Result := eValue;

  if HasMaxValue and (eValue > MaxValue) then
    Result := MaxValue;

  if HasMinValue and (eValue < MinValue) then
    Result := MinValue;
end;

constructor TCustomFloatEdit.Create(AOwner: TComponent);
begin
  inherited;

  { default is right aligned }
  Alignment := taRightJustify;
  Text      := '';

  { by default no min & max }
  feMinValue    := 0.0;
  feMaxValue    := 0.0;
  fbHasMinValue := False;
  fbHasMaxValue := False;
  fiMaxDecimals := FLOAT_EDIT_DEFAULT_MAX_DECIMALS;
end;

function TCustomFloatEdit.GetValue: extended;
begin
  if Text = '' then
    Result := 0.0
  else
    Result := StrToFloatDef(Text, 0.0);
end;

function TCustomFloatEdit.IsInRange(eValue: extended): boolean;
begin
  Result := True;

  if HasMaxValue and (eValue > MaxValue) then
    Result := False;

  if HasMinValue and (eValue < MinValue) then
    Result := False;
end;

function TCustomFloatEdit.IsInRange(sValue: string): boolean;
var
  leValue: extended;
begin
  if sValue = '' then
    leValue := 0.0
  else
    leValue := StrToFloatDef(sValue, 0.0);

  Result := IsInRange(leValue);
end;

procedure TCustomFloatEdit.KeyPress(var key: char);
var
  lsNewText: string;
  iDotPos:   integer;
begin
  { not intersted in control chars }
  if (Ord(Key)) < Ord(' ') then
  begin
    inherited KeyPress(Key);
    exit;
  end;

  { allow only digits, '.' and minus sign }
  if not CharIsNumber(Key) then
    Key := #0;

  if Key = '-' then
  begin
    { are any neg. numbers in range ? }
    if HasMinValue and (MinValue >= 0) then
      Key := #0
    { minus sign only as first char }
    else if SelStart <> 0 then
      Key := #0;

    { only 1 minus sign }
    if StrLeft(lsNewText, 2) = '--' then
      Key := #0;
  end;

  { no leading zeros }
  if (SelStart = 0) and (Key = '0') and (StrLeft(lsNewText, 1) = '0') then
    Key := #0;

  { no leading '-0' }
  if (SelStart = 1) and (Key = '0') and (StrLeft(lsNewText, 2) = '-0') then
    Key := #0;

  iDotPos := Pos('.', Text);

  if Key = '.' then
  begin
    { only one dot }
    if iDotPos > 0 then
      Key := #0
  end;

  { check number of decimal digits }
  if (iDotPos > 0) and (SelStart > iDotPos) and
    ((Length(Text) - iDotPos) >= MaxDecimals) then
  begin
    Key := #0;
  end;

  { disallow the keypress if the value would go out of range }
  lsNewText := GetChangedText(Text, SelStart, SelLength, Key);
  if not IsInRange(lsNewText) then
    Key := #0;

  if Key <> #0 then
    inherited;
end;

procedure TCustomFloatEdit.SetMaxDecimals(iNewValue: integer);
begin
  if iNewValue < 0 then
    iNewValue := 0;
  if iNewValue > FLOAT_EDIT_MAX_MAX_DECIMALS then
    iNewValue := FLOAT_EDIT_MAX_MAX_DECIMALS;

  fiMaxDecimals := iNewValue;
end;

procedure TCustomFloatEdit.SetMaxValue(eNewValue: extended);
begin
  feMaxValue := eNewValue;

  { make min value consistent }
  if MinValue > MaxValue then
    feMinValue := MaxValue;
end;

procedure TCustomFloatEdit.SetMinValue(eNewValue: extended);
begin
  feMinValue := eNewValue;

  { make max value consistent }
  if MaxValue < MaxValue then
    feMaxValue := MinValue;
end;

procedure TCustomFloatEdit.SetValue(eNewValue: extended);
begin
  Text := FloatToStr(ChopToRange(eNewValue));
end;

{-------------------------------------------------------------------------------
  TCustomYearEdit }

procedure TCustomYearEdit.Assign(Source: TPersistent);
var
  lcSource: TCustomYearEdit;
begin
  inherited;

  if Source is TCustomYearEdit then
  begin
    lcSource  := Source as TCustomYearEdit;
    PivotYear := lcSource.PivotYear;
  end;
end;

constructor TCustomYearEdit.Create(AOwner: TComponent);
begin
  inherited;

  { the system var is 'TwoDigitYearCenturyWindow', which by default will have a value of 50
    we then want fiPivotYear to be 1950,
    if current date is 2020, and TwoDigitYearCenturyWindow is 90,
      then fiPivotYear should be 1990
    if current date is 2120, and TwoDigitYearCenturyWindow is 90, (this code may last that long?)
      then fiPivotYear should be 2090.
    It helps that fiPivotYear is assumed to be in the past 100 years
  }
  fiPivotYear := CenturyOfDate(now) + TwoDigitYearCenturyWindow;
  if fiPivotYear > YearOfDate(now) then
    fiPivotYear := fiPivotYear - 100;

  HasMinValue := True;
  MinValue    := 0;
  HasMaxValue := True;
  MaxValue    := 9999;
end;

procedure TCustomYearEdit.DoExit;
begin
  { allow the user to type & on exit, translate any 2 digit years to 4 digits }
  FixYear;
  inherited;
end;

procedure TCustomYearEdit.FixYear;
begin
  Value := MyMakeYear4Digit(Value, PivotYear);
end;

function TCustomYearEdit.GetValue: integer;
var
  li: integer;
begin
  li := inherited GetValue;
  Result := MyMakeYear4Digit(li, PivotYear);
end;

procedure TCustomYearEdit.SetValue(iNewValue: integer);
var
  li: integer;
begin
  li := MyMakeYear4Digit(iNewValue, PivotYear);
  inherited SetValue(li);
end;

{ TCustomCurrencyEdit }

procedure TCustomCurrencyEdit.Assign(Source: TPersistent);
var
  lcSource: TCustomCurrencyEdit;
begin
  inherited;

  if Source is TCustomCurrencyEdit then
  begin
    lcSource    := Source as TCustomCurrencyEdit;
    MinValue    := lcSource.MinValue;
    MaxValue    := lcSource.MaxValue;
    HasMinValue := lcSource.HasMinValue;
    HasMaxValue := lcSource.HasMaxValue;
  end;
end;

function TCustomCurrencyEdit.ChopToRange(eValue: currency): currency;
begin
  Result := eValue;

  if HasMaxValue and (eValue > MaxValue) then
    Result := MaxValue;

  if HasMinValue and (eValue < MinValue) then
    Result := MinValue;
end;

constructor TCustomCurrencyEdit.Create(AOwner: TComponent);
begin
  inherited;

  { default is right aligned }
  Alignment := taRightJustify;
  Text      := '';

  { by default no min & max }
  feMinValue    := 0.0;
  feMaxValue    := 0.0;
  fbHasMinValue := False;
  fbHasMaxValue := False;
end;

function TCustomCurrencyEdit.GetValue: currency;
var
  lsText: string;
begin
  lsText := StripNonNumberChars(Text);

  if lsText = '' then
    Result := 0.0
  else
    Result := StrToCurrDef(lsText, 0.0);
end;

function TCustomCurrencyEdit.IsInRange(eValue: currency): boolean;
begin
  Result := True;

  if HasMaxValue and (eValue > MaxValue) then
    Result := False;

  if HasMinValue and (eValue < MinValue) then
    Result := False;
end;

function TCustomCurrencyEdit.IsInRange(sValue: string): boolean;
var
  leValue: extended;
begin
  if sValue = '' then
    leValue := 0.0
  else
    leValue := StrToCurrDef(sValue, 0.0);

  Result := IsInRange(leValue);
end;

procedure TCustomCurrencyEdit.KeyPress(var key: char);
var
  lsNewText: string;
  iDotPos:   integer;
begin
  { not intersted in control chars }
  if (Ord(Key)) < Ord(' ') then
  begin
    inherited KeyPress(Key);
    exit;
  end;

  { allow only digits, '.' and minus sign }
  if not CharIsNumber(Key) then
    Key := #0;

  if Key = '-' then
  begin
    { are any neg. numbers in range ? }
    if HasMinValue and (MinValue >= 0) then
      Key := #0
    { minus sign only as first char }
    else if SelStart <> 0 then
      Key := #0;

    { only 1 minus sign }
    if StrLeft(lsNewText, 2) = '--' then
      Key := #0;
  end;

  { no leading zeros }
  if (SelStart = 0) and (Key = '0') and (StrLeft(lsNewText, 1) = '0') then
    Key := #0;

  { no leading '-0' }
  if (SelStart = 1) and (Key = '0') and (StrLeft(lsNewText, 2) = '-0') then
    Key := #0;

  iDotPos := Pos('.', Text);

  if Key = '.' then
  begin
    { only one dot }
    if iDotPos > 0 then
      Key := #0
  end;

  { check number of decimal digits }
  if (iDotPos > 0) and (SelStart > iDotPos) and
    ((Length(Text) - iDotPos) >= CurrencyDecimals) then
  begin
    Key := #0;
  end;

  { disallow the keypress if the value would go out of range }
  lsNewText := GetChangedText(Text, SelStart, SelLength, Key);
  if not IsInRange(lsNewText) then
    Key := #0;

  if Key <> #0 then
    inherited;
end;

procedure TCustomCurrencyEdit.SetMaxValue(eNewValue: currency);
begin
  feMaxValue := eNewValue;

  { make min value consistent }
  if MinValue > MaxValue then
    feMinValue := MaxValue;
end;

procedure TCustomCurrencyEdit.SetMinValue(eNewValue: currency);
begin
  feMinValue := eNewValue;

  { make max value consistent }
  if MaxValue < MaxValue then
    feMaxValue := MinValue;
end;

procedure TCustomCurrencyEdit.SetValue(eNewValue: currency);
begin
  Text := CurrToStr(ChopToRange(eNewValue));
end;

{-------------------------------------------------------------------------------
 TFormattedCurrencyEdit }

procedure TCustomFormattedCurrencyEdit.DoEnter;
begin
  UnFormatText;
  inherited;
end;

procedure TCustomFormattedCurrencyEdit.DoExit;
begin
  inherited;
  FormatText;
end;

procedure TCustomFormattedCurrencyEdit.FormatText;
const
  MONEY_FORMAT = '%m'; // see help on "format strings"
begin
  Text := Format(MONEY_FORMAT, [Value]);
end;

{ this sets the text to then non-formatted value, the reads it, then
  sets it to the formatted value. This could be improved }
procedure TCustomFormattedCurrencyEdit.SetValue(eNewValue: currency);
begin
  inherited;
  FormatText;
 { what if the control has focus? It won't if this occurs as the result of a button click }
end;

procedure TCustomFormattedCurrencyEdit.UnformatText;
begin
  Text := StripNonNumberChars(Text);
end;


end.