{(*}
(*------------------------------------------------------------------------------
 Delphi Code formatter source code 

The Original Code is Converter.pas, released April 2000.
The Initial Developer of the Original Code is Anthony Steele. 
Portions created by Anthony Steele are Copyright (C) 1999-2000 Anthony Steele.
All Rights Reserved. 
Contributor(s): Anthony Steele. 

The contents of this file are subject to the Mozilla Public License Version 1.1
(the "License"). you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://www.mozilla.org/NPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied.
See the License for the specific language governing rights and limitations 
under the License.
------------------------------------------------------------------------------*)
{*)}

unit Converter;

{ AFS 28 Nov 1999
  I don't want the controlling logic to be in a form
  so this class has it
}


interface

uses
    { delphi } Classes,
    { local } ConvertTypes, Reader, Writer, Tokeniser, Log,
  TokenProcessPipeline, Settings;

type

  TConverter = class(TObject)
  private
    fcReader: TReader;
    fcWriter: TWriter;
    fcTokeniser: TTokeniser;
    fiCount: integer;

    fcProcess: TTokenProcessPipeline;
    fcLog: TLog;
    fcSettings: TSettings;

    fbAbort, fbYesAll: boolean;

    fsOriginalFileName: string;

    fOnStatusMessage: TStatusMessageProc;

    procedure SendStatusMessage(const ps: string);

    procedure GetFileNames(const psDir: string; psFiles: TStringList);
    procedure GetDirNames(const psDir: string; psFiles: TStringList);

    procedure Flush;
    procedure SetSettings(const Value: TSettings);

    procedure FinalSummary;
    procedure OnFileStart;
    procedure OnFileEnd;

  protected
  public
    constructor Create;
    destructor Destroy; override;

    procedure ProcessDirectory(const psDir: string);
    procedure ProcessFile(const psInput: string);
    procedure Convert;
    function OriginalFileName: string;

    property Settings: TSettings read fcSettings write SetSettings;

    property OnStatusMessage: TStatusMessageProc read fOnStatusMessage write
      fOnStatusMessage;

  end;

implementation

uses
    { delphi } Windows, SysUtils, Dialogs, Controls, FileCtrl,
    { jcl } JclFileUtils,
    { local } SetLog;

{ TConverter }

function FileIsReadOnly(const ps: string): boolean;
var
  liAttr: integer;
begin
  liAttr := FileGetAttr(ps);
  Result := ((liAttr and faReadOnly) <> 0);
end;

procedure TConverter.ProcessFile(const psInput: string);
var
  lsMessage, lsOut, lsTemp: string;
  wRes:                     word;
begin
  if psInput = '' then
  begin
    ShowMessage('Select a file');
    exit;
  end;

  if not FileExists(psInput) then
  begin
    ShowMessage('The file "' + psInput + '" does not exist');
    exit;
  end;


  if (Settings.FileSettings.SourceMode <> fmSingleFile) then
  begin
    lsTemp := PathExtractFileNameNoExt(psInput);
    if Settings.FileSettings.FileIsExcluded(lsTemp) then
    begin
      fcLog.LogMessage('Exluded file: ' + psInput);
      exit;
    end;
  end;

  { all kinds of chaos ensues if you work with readonly files,
    for e.g. you can rename them to .bak, but on the next run you will be unable to delete the old .bak files.
    It is only safe when the source is read not written, ie "output to seperate file" backup mode
  }

  if (Settings.FileSettings.BackupMode <> cmSeperateOutput) and
    (FileIsReadOnly(psInput)) then
  begin
    fcLog.LogError('File: ' + psInput + ' cannot be processed as it is read only');
    exit;
  end;

  lsMessage := 'Processing file ' + psInput;
  if Settings.Log.LogLevel in [eLogFiles, eLogTokens] then
    fcLog.LogMessage(lsMessage);
  SendStatusMessage(lsMessage);
  fsOriginalFileName := psInput;

  lsOut := Settings.FileSettings.GetOutputFileName(psInput);

  if Settings.FileSettings.BackupMode <> cmInplace then
  begin
    if lsOut = '' then
    begin
      ShowMessage('No output/backup file specifed');
      exit;
    end;

    if FileExists(lsOut) then
    begin
      if fbYesAll then
        wRes := mrYes
      else
        wRes := MessageDlg('Output/backup file ' + lsOut + ' exists already. Remove it?',
          mtConfirmation, [mbYes, mbNo, mbAll, mbAbort], 0);

      if wRes = mrAll then
      begin
        fbYesAll := True;
        wRes     := mrYes;
      end;

      if wRes = mrYes then
      begin
        if not DeleteFile(lsOut) then
          raise Exception.Create('TConverter.ProcessFile: ' +
            'Failed to delete file ' + lsOut);
      end
      else if wRes = mrNo then
      begin
        exit;
      end
      else if wRes = mrAbort then
      begin
        fbAbort := True;
        exit;
      end;
    end;
  end;

  if Settings.FileSettings.BackupMode = cmInPlace then
  begin
    { rename the original file to an arbitrary temp file,
      write processed code back to the original file,
      delete the temp }


    lsTemp := FileGetTempName('Con');
    if FileExists(lsTemp) then
      if not DeleteFile(lsTemp) then
        raise Exception.Create('TConverter.ProcessFile: ' +
          'Failed to delete temp file ' + lsTemp);

    if not RenameFile(psInput, lsTemp) then
      raise Exception.Create('TConverter.ProcessFile: ' +
        'could not rename source file ' + psInput + ' to ' + lsTemp);

    fcReader.SourceFileName := lsTemp;
    fcWriter.OutputFileName := psInput;
  end
  else if Settings.FileSettings.BackupMode = cmInPlaceWithBackup then
  begin
    { rename the original file to the backup file name,
      write processed code back to the original file }


    if not RenameFile(psInput, lsOut) then
      raise Exception.Create('TConverter.ProcessFile: ' +
        'could not rename source file ' + psInput + ' to ' + lsOut);

    fcReader.SourceFileName := lsOut;
    fcWriter.OutputFileName := psInput;
  end
  else
  begin
    { simple. Source is source, dest is dest }

    fcReader.SourceFileName := psInput;
    fcWriter.OutputFileName := lsOut;
    if FileExists(lsOut) then
      raise Exception.Create('TConverter.ProcessFile: ' +
        'Destination file ' + lsOut + ' exists already');
  end;

  OnFileStart;
  fcWriter.WriteAll;
  Flush;
  OnFileEnd;

  Inc(fiCount);

  if Settings.FileSettings.BackupMode = cmInPlace then
    if not DeleteFile(lsTemp) then
      fcLog.LogError('TConverter.ProcessFile: ' +
        'Failed to delete temp file ' + lsTemp + ' for ' + psInput);
end;

procedure TConverter.ProcessDirectory(const psDir: string);
var
  lsMessage: string;
  lsNames:   TStringList;
  lsDir:     string;
  liLoop:    integer;
begin
  if not DirectoryExists(psDir) then
  begin
    ShowMessage('The directory ' + psDir + ' does not exist');
    exit;
  end;

  lsDir := PathAddSeparator(psDir);

  lsMessage := 'Processing directory ' + lsDir;
  if Settings.Log.LogLevel in [eLogFiles, eLogTokens] then
    fcLog.LogMessage(lsMessage);
  SendStatusMessage(lsMessage);

  lsNames := TStringList.Create;
  try { finally free }
    GetFileNames(lsDir, lsNames);

    for liLoop := 0 to lsNames.Count - 1 do
    begin
      ProcessFile(lsDir + lsNames[liLoop]);
      if fbAbort then
        break;
    end;

    { all subdirs }
    if Settings.FileSettings.SourceMode = fmDirectoryRecursive then
    begin
      lsNames.Clear;
      GetDirNames(lsDir, lsNames);

      for liLoop := 0 to lsNames.Count - 1 do
      begin
        ProcessDirectory(lsDir + lsNames[liLoop]);
        if fbAbort then
          break;
      end;
    end;

  finally
    lsNames.Free;
  end;
end;

procedure TConverter.GetFileNames(const psDir: string; psFiles: TStringList);
var
  rSearch: TSearchRec;
  lsName, lsExt, lsSearch: string;
  bDone:   boolean;
begin
  Assert(psDir <> '');
  Assert(psFiles <> nil);

  { for all pas files in the dir }
  lsSearch := psDir + '*.*';
  FillChar(rSearch, Sizeof(TSearchRec), 0);
  bDone := (FindFirst(lsSearch, 0, rSearch) <> 0);

  while not bDone do
  begin
    lsName := rSearch.Name;
    Assert(lsName <> '');
    if (rSearch.Attr and faDirectory > 0) then
      continue;

    lsExt := ExtractFileExt(lsName);
    if (AnsiCompareText(lsExt, '.pas') = 0) or (AnsiCompareText(lsExt, '.dpr') = 0) then
      psFiles.Add(lsName);

    bDone := (FindNext(rSearch) <> 0);
    Assert(bDone or (rSearch.Name <> lsName));
  end;
  FindClose(rSearch);
end;

procedure TConverter.GetDirNames(const psDir: string; psFiles: TStringList);
var
  rSearch:  TSearchRec;
  lsSearch: string;
  bDone:    boolean;
begin
  Assert(psDir <> '');
  Assert(psFiles <> nil);

  lsSearch := psDir + '*.*';
  FillChar(rSearch, Sizeof(TSearchRec), 0);
  bDone := (FindFirst(lsSearch, faDirectory, rSearch) <> 0);

  while not bDone do
  begin
    if (rSearch.Attr and faDirectory > 0) and
      (rSearch.Name <> '.') and (rSearch.Name <> '..') then
      psFiles.Add(rSearch.Name);

    bDone := (FindNext(rSearch) <> 0);
  end;
  FindClose(rSearch);
end;

constructor TConverter.Create;
begin
  inherited;

  { create owned objects }
  fcReader    := TReader.Create;
  fcWriter    := TWriter.Create;
  fcTokeniser := TTokeniser.Create;
  fcLog       := TLog.Create;

  fcProcess     := TTokenProcessPipeline.Create;
  fcProcess.Log := fcLog;

  { events }
  fcTokeniser.GetFileName := OriginalFileName;
  fcProcess.GetFileName   := OriginalFileName;

  { wire them together }
  fcTokeniser.Reader := fcReader;
  fcTokeniser.Log    := fcLog;

  fcProcess.Source := fcTokeniser;
  fcWriter.Source  := fcProcess;
end;

destructor TConverter.Destroy;
begin
  FreeAndNil(fcReader);
  FreeAndNil(fcWriter);
  FreeAndNil(fcTokeniser);
  FreeAndNil(fcLog);
  FreeAndNil(fcProcess);

  inherited;
end;


procedure TConverter.Convert;
var
  dwStart, dwElapsed: DWord;
begin
  if Settings.Log.LogTime then
    dwStart := GetTickCount
  else
    dwStart := 0;

  fbAbort  := False;
  fbYesAll := False;
  fiCount  := 0;

  { all processors must chekc thier inclusion settings
    as this may have changed from the UI }

  fcProcess.SetIsEnabled;

  { process file(s) }
  case Settings.FileSettings.SourceMode of
    fmSingleFile:
      ProcessFile(Settings.FileSettings.Input);
    fmDirectory, fmDirectoryRecursive:
      ProcessDirectory(Settings.FileSettings.Input);
    else
      raise Exception.Create('TConverter.Convert: Bad file recurse type');
  end;


  if Settings.Log.LogTime then
  begin
    dwElapsed := GetTickCount - dwStart;
    fcLog.LogPriorityMessage('Run took ' + FloatToStr(dwElapsed / 1000) + ' seconds')
  end;

  FinalSummary;
  fcLog.CloseLog;

  if Settings.Log.ViewLogAfterRun then
    Settings.Log.ViewLog;
end;

procedure TConverter.SendStatusMessage(const ps: string);
begin
  if Assigned(fOnStatusMessage) then
    fOnStatusMessage(ps);
end;

procedure TConverter.Flush;
begin
  fcReader.Clear;
  fcProcess.Flush;
  fcWriter.Close;
end;

procedure TConverter.SetSettings(const Value: TSettings);
begin
  fcSettings := Value;

  { tell the owned objects about it }
  fcProcess.Settings   := Value;
  fcTokeniser.Settings := Value;
  fcLog.Settings       := Value;
end;

procedure TConverter.FinalSummary;
var
  lsMessage: string;
begin
  if fbAbort then
    lsMessage := 'Aborted after ' + IntToStr(fiCount) + ' files'
  else
    lsMessage := 'Finished processing ' + IntToStr(fiCount) + ' files';

  SendStatusMessage(lsMessage);

  fcTokeniser.FinalSummary;
  fcProcess.FinalSummary;

  fcLog.EmptyLine;
  fcLog.LogPriorityMessage(lsMessage);
end;

procedure TConverter.OnFileStart;
begin
  fcTokeniser.OnFileStart;
  fcProcess.OnFileStart;
end;


procedure TConverter.OnFileEnd;
begin
  fcTokeniser.OnFileEnd;
  fcProcess.OnFileEnd;
end;

function TConverter.OriginalFileName: string;
begin
  Result := fsOriginalFileName;
end;

end.