{(*}
(*------------------------------------------------------------------------------
 Delphi Code formatter source code 

The Original Code is Reader.pas, released April 2000.
The Initial Developer of the Original Code is Anthony Steele. 
Portions created by Anthony Steele are Copyright (C) 1999-2000 Anthony Steele.
All Rights Reserved. 
Contributor(s): Anthony Steele. 

The contents of this file are subject to the Mozilla Public License Version 1.1
(the "License"). you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://www.mozilla.org/NPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied.
See the License for the specific language governing rights and limitations
under the License.
------------------------------------------------------------------------------*)
{*)}

unit Reader;

{ Created AFS 27 November 1999
  reader for Code formatting util

  The method is to first read the entire file into a string
  This is a textbook optimization - 1 read for the whole file
  instead of 1 per char. The file may be large
  (the largest file that ships with Delphi5, excel2000.pas, is 4Mb!!!!)
  but even this should fit into memory
  This technique is not optimised for files of that size,
  but hey, that code has got to be machine-generated anyway.
  Why would it need machine-reformatting?
}

interface

type
  TReader = class(TObject)
  private
    { working vars }
    fsSource: string;
    fiSourceLength: integer;

    fiReadIndex, fiBufferLength: integer;
    fbHasRead: boolean;

    { property implementation }
    FsSourceFileName: string;

    procedure SetSourceFileName(const psValue: string);
    function GetBuffer: string;

    procedure SetBufferLength(const piValue: integer);

    { worker procs }
    procedure ReadFile;
  protected
  public
    constructor Create;

    procedure Clear;

    procedure Consume; overload;
    procedure Consume(piChars: integer); overload;
    function ConsumeBuffer: string;

    function Current: char;
    function Last: char;

    procedure IncBuffer;
    procedure DecBuffer;

    function BufferCharsLeft(const piLength: integer): string;
    function BufferCharsRight(const piLength: integer): string;

    function EndOfFile: boolean;
    function BufferEndOfFile: boolean;

    property SourceFileName: string read FsSourceFileName write SetSourceFileName;

    property Buffer: string read GetBuffer;
    property BufferLength: integer read fiBufferLength write SetBufferLength;
  end;

implementation

uses
    {delphi } SysUtils,
    { local } VCLFunctions;

{ TReader }

constructor TReader.Create;
begin
  inherited;

  fsSourceFileName := '';

  Clear;
end;

procedure TReader.Clear;
begin
  fsSource       := '';
  fiSourceLength := 0;

  fiReadIndex    := -1;
  fiBufferLength := 1;
  fbHasRead      := False;

  FsSourceFileName := '';
end;

procedure TReader.SetSourceFileName(const psValue: string);
begin
  FsSourceFileName := psValue;
end;

procedure TReader.ReadFile;
var
  lfSource: file;
  liSize, liAmtTransfered: integer;
  pChars:   Pointer;
begin
  if fbHasRead then
    exit;

  // Open the file
  Assert(FileExists(SourceFileName));
  FileMode := 0; // this bizarre var changes the mode that reset uses to read only
  AssignFile(lfSource, SourceFileName);
  // 1 byte per "record". sheesh, this stuff is neolithic. Give us a TFile object awready!
  Reset(lfSource, 1);

  // make room in the string
  liSize := FileSize(lfSource);
  Assert(liSize > 0);
  SetLength(fsSource, liSize);
  pChars := PChar(fsSource);
  FillChar(pChars^, liSize, #0);

  // read it all in one fell gulp
  BlockRead(lfSource, pChars^, liSize, liAmtTransfered);
  Assert(liAmtTransfered = liSize);

  CloseFile(lfSource);

  fiSourceLength := Length(fsSource);
  Assert(fiSourceLength = liSize);

  fiReadIndex    := 1;
  fiBufferLength := 1;
  fbHasRead      := True;
end;

function TReader.EndOfFile: boolean;
begin
  Result := (fiReadIndex > fiSourceLength);
end;

function TReader.BufferEndOfFile: boolean;
begin
  Result := ((fiReadIndex + BufferLength - 1) >= fiSourceLength);
end;

{ make sure that the buffer is at least this length }
procedure TReader.SetBufferLength(const piValue: integer);
begin
  if not fbHasRead then
    ReadFile;

  Assert(piValue >= 1);
  fiBufferLength := piValue;
end;

procedure TReader.IncBuffer;
begin
  inc(fiBufferLength);

  { chop to the remaining length }
  if (fiReadIndex + fiBufferLength - 1) > fiSourceLength then
    fiBufferLength := fiSourceLength - fiReadIndex + 1;
end;

procedure TReader.DecBuffer;
begin
  dec(fiBufferLength);
  if fiBufferLength < 1 then
    fiBufferLength := 1;
end;


procedure TReader.Consume;
begin
  inc(fiReadIndex);
end;

procedure TReader.Consume(piChars: integer);
begin
  Assert(piChars > 0);
  inc(fiReadIndex, piChars);
  dec(fiBufferLength, piChars);
  if fiBufferLength < 1 then
    fiBufferLength := 1;
end;

function TReader.ConsumeBuffer: string;
begin
  Result := Buffer;
  Consume(BufferLength);
  BufferLength := 1;
end;

function TReader.GetBuffer: string;
begin
  Result := Copy(fsSource, fiReadIndex, BufferLength);
end;

function TReader.BufferCharsLeft(const piLength: integer): string;
begin
  Result := '';
  if fsSource = '' then
    exit;

  Result := Copy(fsSource, fiReadIndex, piLength);
end;

function TReader.BufferCharsRight(const piLength: integer): string;
var
  liStart: integer;
begin
  Result := '';
  if fsSource = '' then
    exit;
  Assert(piLength <= BufferLength);

  liStart := (fiReadIndex + BufferLength - piLength);
  if liStart > fiSourceLength then
    Result := #0
  else
    Result := Copy(fsSource, liStart, piLength);
end;

function TReader.Current: char;
begin
  if not fbHasRead then
    ReadFile;

  if (fiReadIndex > fiSourceLength) then
    Result := #0
  else
    Result := fsSource[fiReadIndex];
end;

function TReader.Last: char;
var
  liIndex: integer;
begin
  liIndex := fiReadIndex + BufferLength - 1;

  if BufferLength < 0 then
    Result := #0
  else if (liIndex > fiSourceLength) or (liIndex < 1) then
    Result := #0
  else
    Result := fsSource[liIndex];
end;

end.