{(*}
(*------------------------------------------------------------------------------
 Delphi Code formatter source code 

The Original Code is Writer.pas, released April 2000.
The Initial Developer of the Original Code is Anthony Steele. 
Portions created by Anthony Steele are Copyright (C) 1999-2000 Anthony Steele.
All Rights Reserved. 
Contributor(s): Anthony Steele. 

The contents of this file are subject to the Mozilla Public License Version 1.1
(the "License"). you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://www.mozilla.org/NPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied.
See the License for the specific language governing rights and limitations 
under the License.
------------------------------------------------------------------------------*)
{*)}

unit Writer;

{ AFS 28 November 1999
  Writer - final output stage of code formattter

  AFS 22 July 2K - optimised by using a string to store tokens,
  and writing the file at once
  }

interface

uses TokenType, Token, TokenSource;

type
  TWriter = class(TObject)
  private
    { working vars }
    fbBOF: boolean;
    fsDestText: string;
    fiTokensWritten: integer;


    { properties }
    FOutputFileName: string;
    FTokenSource: TTokenSource;
    procedure SetOutputFileName(const Value: string);

    { worker procs }
    procedure SetTokenSource(const Value: TTokenSource);

  protected

  public
    constructor Create;
    destructor Destroy; override;

    procedure Close;

    procedure WriteOut(const ch: char); overload;
    procedure WriteOut(const st: string); overload;
    procedure WriteOut(const pt: TToken); overload;

    { return True if there is more }
    function DoNextToken: boolean;
    procedure WriteAll;

    property Source: TTokenSource read FTokenSource write SetTokenSource;

    property OutputFileName: string read FOutputFileName write SetOutputFileName;
    property BOF: boolean read fbBOF;
  end;

implementation
                               
uses { delphi } SysUtils;

const
  MAX_TOKENS = 100000;

  { TWriter }

constructor TWriter.Create;
begin
  inherited;
  FOutputFileName := '';
  fbBOF           := True;
  FTokenSource    := nil;
end;

destructor TWriter.Destroy;
begin
  Close;
  inherited;
end;

procedure TWriter.SetOutputFileName(const Value: string);
begin
  FOutputFileName := Value;
end;

procedure TWriter.WriteOut(const ch: char);
begin
  fsDestText := fsDestText + ch;
  fbBOF      := False;
end;

procedure TWriter.WriteOut(const st: string);
begin
  //Assert(st <> #0);
  if st = #0 then
    exit;

  fsDestText := fsDestText + st;
  fbBOF      := False;
end;

procedure TWriter.SetTokenSource(const Value: TTokenSource);
begin
  FTokenSource := Value;
end;

procedure TWriter.WriteOut(const pt: TToken);
begin
  WriteOut(pt.SourceCode);
  inc(fiTokensWritten);
end;

function TWriter.DoNextToken: boolean;
var
  lcToken: TToken;
begin
  lcToken := Source.GetNextToken;
  Assert(lcToken <> nil);

  if lcToken.TokenType <> ttEOF then
    WriteOut(lcToken);

  Result := not (lcToken.TokenType = ttEOF);
  lcToken.Free;

  { this can also be the result of parsing
    machine generated TLB.pas files
    this is piointless, give it up. }
  if fiTokensWritten > MAX_TOKENS then
    raise Exception.Create('Too many tokens (' + IntToStr(MAX_TOKENS) +
      ') in file ' + OutputFileName +
      ' Suspected program error. ' +
      'If you are parsing large input files, raise the limit.');
end;

procedure TWriter.WriteAll;
var
  lbMore: boolean;
begin
  Assert(Source <> nil);
  lbMore := True;
  fiTokensWritten := 0;

  while lbMore do
    lbMore := DoNextToken;
end;

procedure TWriter.Close;
var
  lfOutput: file;
  pChars:   Pointer;
begin
  if BOF then
    exit;

  Assert(OutputFileName <> '');
  Assert(not FileExists(OutputFileName));
  pChars := PChar(fsDestText);

  { write the file }
  AssignFile(lfOutput, OutputFileName);
  Rewrite(lfOutput, 1);
  BlockWrite(lfOutput, pChars^, Length(fsDestText));
  CloseFile(lfOutput);

  { reste state }
  FOutputFileName := '';
  fsDestText      := '';
  fbBOF           := True;
end;

end.