{(*}
(*------------------------------------------------------------------------------
 Delphi Code formatter source code 

The Original Code is TokenBareIndent.pas, released April 2000.
The Initial Developer of the Original Code is Anthony Steele. 
Portions created by Anthony Steele are Copyright (C) 1999-2000 Anthony Steele.
All Rights Reserved. 
Contributor(s): Anthony Steele. 

The contents of this file are subject to the Mozilla Public License Version 1.1
(the "License"). you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://www.mozilla.org/NPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied.
See the License for the specific language governing rights and limitations 
under the License.
------------------------------------------------------------------------------*)
{*)}

unit TokenBareIndent;

{ AFS 1 March  2K
 This class introduced to recognse indent on "bare" statements
 ie the statement is not a block but still needs an extra indent

 see the file Testcases\TestLayoutBare.pas for examples of code that excercises this }


interface


uses TokenSource, Token, TokenType;

type

  TTokenBareIndent = class(TBufferedTokenProcessor)
  private
    fiStartIndent: integer;

    function BareStatementStart(const pt: TToken): boolean;
    function BareStatementEnd(const ptStart: TToken): integer;

  protected
    function IsTokenInContext(const pt: TToken): boolean; override;

    function OnProcessToken(const pt: TToken): TToken; override;
  public
    procedure OnFileStart; override;


  end;

implementation

uses WordMap;

{ TTokenBareIndent }

function TTokenBareIndent.BareStatementEnd(const ptStart: TToken): integer;
const
  BareStatementEndToken: TTokenTypeSet = [ttSemiColon];
var
  liCount: integer;
  lt:      TToken;
  lbIf, lbFound: boolean;
  liIfCount: integer;
  lsEndWords: TWordSet;
begin
  { look ahead to find where this bare statement ends }
  liCount := 0;
  lbFound := False;

  lsEndWords := [wEnd];
  if ptStart.TokenType <> ttColon then
    lsEndWords := lsEndWords + [wElse];

  { if Statement processing

    sometimes I think it would have been simpler to use a tree-parsing algorythym
   you can't just scan forward from the then for the first else
   because the else binds to the innermost if
   So must keep a count (liIfCount) of the if/else pairs encountered whilst looking for the end

   if a > 10 then // start here
     if b > 10 then
       Dosomething
     else // don't match here
       DoSomethingElse
   else // match here !
    DoAnotherThing;
  }

  lbIf      := (ptStart.Word in [wThen, wElse]);
  liIfCount := 0;

  lt := BufferTokens(0);
  fiStartIndent := lt.NestingLevel;

  while not lbFound do
  begin
    lt := BufferTokens(liCount);
    inc(liCount);

    { EOF?!? look no further }
    if lt.TokenType = ttEOF then
      lbFound := True;

    { count the ifs }
    if lbIf and (lt.NestingLevel = fiStartIndent) then
    begin
      if (lt.Word = wIf) then
        inc(liIfCount);

      { if ends with an else or ; recognise semicolon before, else after
        dunno why, just works this way }
      if (lt.TokenType = ttSemiColon) then
        Dec(liIfCount);
    end;

    { found it ! }
    if ((lt.TokenType in BareStatementEndToken) or (lt.Word in lsEndWords)) and
      (lt.NestingLevel <= fiStartIndent) and (liIfCount <= 0) then
      lbFound := True;

    if lbIf and (lt.Word = wElse) and (lt.NestingLevel = fiStartIndent) then
      Dec(liIfCount);

    { don't go past the proc end }
    if (lt.ProcedureSection = psNotInProcedure) then
      lbFound := True;
  end;

  { ie stop indenting just before the else or ;
     this would be (liCount - 1),
    but liCount already is the index of the token afer that  }

  Result := liCount - 2;
end;

function IsCaseColon(const pt: TToken): boolean;
begin
  Result := False;

  if pt.TokenType <> ttColon then
    exit;

  if pt.ProcedureSection <> psProcedureBody then
    exit;

  if pt.BracketLevel > 0 then
    exit;

  { !! check if we are in a case statement }
  Result := True;
end;

function TTokenBareIndent.BareStatementStart(const pt: TToken): boolean;
const
  BareStatementStart: TWordSet = [wThen, wElse, wDo];
var
  lcNext:  TToken;
  liIndex: integer;
begin
  Result := False;

  if (pt.ProcedureSection <> psProcedureBody) then
    exit;

  if (pt.Word in BareStatementStart) or IsCaseColon(pt) then
  begin
    liIndex := 0;
    lcNext  := GetBufferTokenWithExclusions(liIndex, [ttWhiteSpace, ttReturn, ttComment]);
    if lcNext.Word <> wBegin then
      Result := True;

    { else if - don't indent the else clause, the if .. then takes care of that }
    if Result and (pt.Word = wElse) and (lcNext.Word = wIf) then
      Result := False;
  end;
end;

function TTokenBareIndent.IsTokenInContext(const pt: TToken): boolean;
begin
  Result := (pt.ProcedureSection = psProcedureBody);
end;

procedure TTokenBareIndent.OnFileStart;
begin
  inherited;
  fiStartIndent := 0;
end;

function TTokenBareIndent.OnProcessToken(const pt: TToken): TToken;
var
  lt: TToken;
  liLoop, liEnd: integer;
begin
  Result := pt;

  { intent a bare block }
  if BareStatementStart(pt) then
  begin
    liEnd := BareStatementEnd(pt);

    for liLoop := 0 to liEnd do
    begin
      lt := BufferTokens(liLoop);
      lt.NestingLevel := lt.NestingLevel + 1;
    end;

    { resume on the next one,
      as the bare statmement can contain bare statements
      e.g. nested ifs }
    fiResume := pt.TokenIndex + 1;
  end;
end;

end.