{(*}
(*------------------------------------------------------------------------------
 Delphi Code formatter source code

The Original Code is TokenContext.pas, released April 2000.
The Initial Developer of the Original Code is Anthony Steele. 
Portions created by Anthony Steele are Copyright (C) 1999-2000 Anthony Steele.
All Rights Reserved. 
Contributor(s): Anthony Steele. 

The contents of this file are subject to the Mozilla Public License Version 1.1
(the "License"). you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://www.mozilla.org/NPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied.
See the License for the specific language governing rights and limitations 
under the License.
------------------------------------------------------------------------------*)
{*)}

unit TokenContext;

{ AFS 23 Dec 1999
 Describe the whereabouts of the token
 and put this data on each token
 so all processors will know where it came from

 }

interface

uses TokenType, Token, TokenSource, WordMap;

type

  TTokenContext = class(TBufferedTokenProcessor)
  private
    feFileSection: TFileSection;
    feProcedureSection: TProcedureSection;
    feDeclarationSection: TDeclarationSection;
    feClassDefinitionSection: TClassDefinitionSection;
    feStructuredType: TStructuredType;

    fiNestingLevel, fiProcedureNestingLevel, fiBracketLevel,
    fiSquareBracketLevel: integer;
    fiCount: integer;

    bInUsesClause: boolean;
    fbRHSEquals: boolean;
    fbRHSAssign: boolean;
    fbRHSColon: boolean;
    fbRHSIn: boolean;
    feRHSofExpressionWord: TWord;

    fbRHSOperand: boolean;
    fbASMBlock: boolean;

    fbInPropertyDefinition: boolean;
    fbInExports: boolean;
    fbDontFormat: boolean;
    { true in class function/proc def&body can be used to distinguish the meaning of the word 'class' }

    fbClassFunction: boolean;

    { true in initialization, finalization & program/lib main body.
      These are code blocks that are like procs }

    fbImplicitBlock: boolean;

    fbForward: boolean;
    // true if the 'forward' directive is found, meaning that this procedure header has no body

    { in a recrd type def, a count of the number of times the word 'case'
     has been encountered }
    fiRecordCases: integer;

    { checks before token gets context }
    procedure CheckInUsesClause(const pt: TToken);
    procedure CheckFileSection(const pt: TToken);
    procedure CheckProcedureSection(const pt: TToken);
    procedure CheckDeclarationSection(const pt: TToken);
    procedure CheckClassDefinitionSection(const pt: TToken);
    procedure CheckBracketLevel(const pt: TToken);
    procedure CheckASM(const pt: TToken);
    procedure CheckForward(const pt: TToken);

    procedure SetContextOnToken(const pt: TToken);

      { checks after token gets context }
    procedure CheckProcedureSectionAfterToken(const pt: TToken);
    procedure CheckOperand(const pt: TToken);
    procedure CheckClassHeader(const pt: TToken);
    procedure CheckEndRHS(const pt: TToken);
    procedure StartProcNestingLevel(const pt: TToken);
    procedure CheckImplicitBlocks(const pt: TToken);
    procedure CheckNestingLevel(const pt: TToken);
    procedure CheckNoFormat(const pt: TToken);
    procedure CheckRHSExpression(const pt: TToken);
    procedure CheckUsesClauseAfter(const pt: TToken);

    procedure ResetState;
    function FindProcedureForward: boolean;
    function NoBodyToProc: boolean;
    procedure SetImplicitBlock;

  protected
    { this one must be ProcessToken, not OnProcessToken
      as it sets up data needed by OnProcessToken }

    function ProcessToken(pt: TToken): TToken; override;
  public
    procedure OnFileStart; override;

  end;

implementation

uses
    { delphi } SysUtils;

{ TTokenContext }

{-------------------------------------------------------------------------------
  overrides }


procedure TTokenContext.OnFileStart;
begin
  feFileSection := fsBeforeInterface;
  ResetState;
  fiCount      := 0;
  fbDontFormat := False;
  fbClassFunction := False;
end;


function TTokenContext.ProcessToken(pt: TToken): TToken;
begin
  // things to do before setting context on this token
  CheckInUsesClause(pt);
  CheckFileSection(pt);
  CheckProcedureSection(pt);
  CheckDeclarationSection(pt);
  CheckClassDefinitionSection(pt);
  CheckBracketLevel(pt);
  CheckASM(pt);
  CheckForward(pt);

  SetContextOnToken(pt);

  {things to do after setting up the token
  i.e. the effect that this token has on subsequent tokens }

  CheckProcedureSectionAfterToken(pt);
  CheckOperand(pt);
  CheckNestingLevel(pt);
  CheckEndRHS(pt);
  CheckClassHeader(pt);
  CheckNoFormat(pt);
  StartProcNestingLevel(pt);
  CheckRHSExpression(pt);
  CheckImplicitBlocks(pt);
  CheckUsesClauseAfter(pt);

  inc(fiCount);
  Result := pt;
end;

{-------------------------------------------------------------------------------
  worker procs }


procedure TTokenContext.ResetState;
begin
  feProcedureSection       := psNotInProcedure;
  feDeclarationSection     := dsNotInDeclaration;
  feClassDefinitionSection := cdNotInClassDefinition;
  feStructuredType         := stNotInStructuredType;

  fiNestingLevel       := 0;
  fiProcedureNestingLevel := 0;
  fiBracketLevel       := 0;
  fiSquareBracketLevel := 0;

  bInUsesClause := False;
  fbRHSEquals   := False;
  fbRHSAssign   := False;
  feRHSofExpressionWord := wUnknown;
  fbRHSOperand  := False;

  fbASMBlock := False;

  fbInExports     := False;
  fbClassFunction := False;
  fbImplicitBlock := False;

  fbForward := False;
end;

{ take a FFworward through the tokens to come
  & look for the 'Forward' Directive
The difficulty will be in knowing when to stop - we cannot depend on context :0 }

function TTokenContext.FindProcedureForward: boolean;
const
  // words that end a proc declaration. Any more?
  EndingWords: TWordSet = [// these are expected
    wBegin, wConst, wResourceString, wVar, wThreadVar, wType, wForward,
    // these will stop us if we run on too far
    wProcedure, wFunction, wConstructor, wDestructor, wEnd];
var
  lt:    TToken;
  bDone: boolean;
  liIndex, liBracketLevel: integer;
begin
  Result := False;

  liIndex := 0;
  liBracketlevel := 0;
  bDone   := False;

  while not bDone do
  begin
    lt := BufferTokens(liIndex);
    inc(liIndex);

    { stop trying on an unexpected EOF }
    if lt.TokenType = ttEOF then
      bDone := True;

    { keep track of brackets to ignore procedure params -
      var & const in here do not denote the end of the proc header }

    if lt.TokenType = ttOpenBracket then
      inc(liBracketLevel);

    if liBracketLevel > 0 then
    begin
      { ignore it all until the close bracket }
      if lt.TokenType = ttCloseBracket then
        dec(liBracketlevel);
      continue;
    end;

    if lt.Word = wForward then
    begin
      Result := True;
      bDone  := True;
    end;

    if lt.Word in EndingWords then
      bDone := True;
  end; { while loop }
end;

procedure TTokenContext.SetImplicitBlock;
begin
  fbImplicitBlock    := True;
  feProcedureSection := psProcedureBody;
  fiNestingLevel     := 1;
  fiProcedureNestingLevel := 1;
end;

{  How to tell if there is should be no procedure body?
    - In the interface file section there is no body
    - In a class/interface def there is no body
    - If the there was a 'forward' directive there is no body
    Otherwise, there must be
}

function TTokenContext.NoBodyToProc: boolean;
begin
  Result := ((feFileSection = fsInterface) or
    (feStructuredType <> stNotInStructuredType) or fbForward);
end;

{-------------------------------------------------------------------------------
  checks that happen before the token is given context }


procedure TTokenContext.CheckInUsesClause(const pt: TToken);
begin
  { uses clause stretches from the word 'uses' to the next semicolon }
  if (pt.TokenType = ttReservedWord) and (pt.Word = wUses) then
    bInUsesClause := True;

  { the word 'in' is used in uses clauses }
  if bInUsesClause then
  begin
    if pt.Word = wIn then
      fbRHSIn := True;
  end;
end;

procedure TTokenContext.CheckFileSection(const pt: TToken);
var
  leOldSection: TFileSection;
begin
  if pt.TokenType <> ttReservedWord then
    exit;

  leOldSection := feFileSection;

  { change section on these keywords }
  case pt.Word of
    wInterface:
    begin
      { interface on the RHS of an equals sign is a COM type def not a file section }
      if not fbRHSEquals then
        feFileSection := fsInterface;
    end;
    wImplementation:
      feFileSection := fsImplementation;
    wInitialization:
      feFileSection := fsInitialization;
    wFinalization:
      feFileSection := fsFinalization;
    wProgram:
      feFileSection := fsProgram;
    wLibrary:
      feFileSection := fsLibrary;
    else
      // do nothing;
  end;

  if leOldSection <> feFileSection then
    ResetState;
end;

function InProcedureDirectives(const pt: TToken): boolean;
const
  // you can legally find these inamongst the directives 
  DirectiveTokens: TTokenTypeSet = [ttWhiteSpace, ttReturn, ttComment,
    ttReservedWordDirective, ttSemiColon, ttNumber];
begin
  { not a token or assocated white space & fomatting }
  Result := (pt.TokenType in DirectiveTokens);

  { a directive, but not one for a proc }
  if (pt.TokenType = ttReservedWordDirective) and
    (not (pt.Word in ProcedureDirectives)) then
    Result := False;
end;

procedure TTokenContext.CheckProcedureSection(const pt: TToken);
begin
  if (feProcedureSection = psProcedureMap) and (pt.TokenType = ttSemiColon) then
    feProcedureSection := psNotInProcedure;

  { after procedure name (and dot), can be an equals sign }
  if (feProcedureSection = psProcedureDefinition) and (pt.Word = wEquals) then
    feProcedureSection := psProcedureMap;

  {
    after the directives (ie as soon as there is a non-directive token)
    and not expecting procedure body, end the procedure def.

    note that not all directives are for procedures.
    Particularly "private", "public" et all are not
    e.g. in the case of a class that has the text
    procedure Fred; virtual; public
    'public' is a directive but not part of the proc's directives
    }

  if (feProcedureSection = psProcedureDirectives) and
    (not InProcedureDirectives(pt)) then
  begin
    if NoBodyToProc then
    begin
      feProcedureSection := psNotInProcedure; // end of proc def, no body
      fbClassFunction    := False;
      fbForward          := False;
    end
    else
    begin
      feProcedureSection := psProcedureDeclarations;
    end;
  end;

  { function/procedure definition/header
    stretches from that keyword to the semicolon
    (And not a semicolon in the parameter list in the brackets) }

  if (pt.Word in ProcedureWords) and ((not fbRHSEquals) or
    (feStructuredType in ObjectTypes)) then
  begin
    feProcedureSection := psProcedureDefinition;
  end;

  if (feProcedureSection = psProcedureDefinition) and (pt.TokenType = ttSemiColon) and
    (fiBracketLevel = 0) then
  begin
    feProcedureSection := psProcedureDirectives;
  end;

  if (feProcedureSection = psProcedureDirectives) and (pt.Word in Declarations) then
  begin
    { in the interface section the var & const are not part of the procedure - they are freestanding
     !! must also happen in implementaion classes }

    if (feFileSection = fsImplementation) and
      (feStructuredType = stNotInStructuredType) then
      feProcedureSection := psProcedureDeclarations
    else
    begin
      feProcedureSection := psNotInProcedure;
      fbClassFunction    := False;
    end;
  end;

  { enter a procedure body on the word begin after a procedure header + local consts and vars }

  if (feProcedureSection in [psProcedureDirectives, psProcedureDeclarations]) and
    (feStructuredType = stNotInStructuredType) and
    (pt.Word in [wBegin, wASM]) and (fiNestingLevel = 0) then
  begin
    feProcedureSection   := psProcedureBody;
    feDeclarationSection := dsNotInDeclaration;
  end;

  { Must do this before conext is put onto the token }

  {a misc. test that fits into no greater scheme :(
  catch the construct
     repeat a := a + 1 until CowsComeHome;
  or even
     if fred then bVal = (a or b) else GoHomeInstead;
  or
    for e.g. if a > 12 then b := True else CallAProc;

     RHSAssign and RHSEquals stops with the else/Until/Finally/end keyword,
  }

  if pt.Word in BlockEndWords then
  begin
    fbRHSAssign := False;
    fbRHSEquals := False;
  end;
end;

procedure TTokenContext.CheckDeclarationSection(const pt: TToken);
begin
  if pt.TokenType = ttReservedWord then
  begin
    { change section on these keywords }
    case pt.Word of
      wType:
        feDeclarationSection := dsType;
      wLabel:
        feDeclarationSection := dsLabel;
      wConst, wResourceString:
        if fiBracketLevel = 0 then
          feDeclarationSection := dsConst;
      wVar, wThreadVar:
        if fiBracketLevel = 0 then
          feDeclarationSection := dsVar;
      wProcedure, wFunction, wConstructor, wDestructor:
      begin
        { encountering the word function/procedure ends a type/const decalration section
          except when
          - it is on the RHS of an equals sign, e.g type TFredProc = procedure (x:Integer);
          - it is in a class or interface def
        }

        if (not fbRHSEquals) and (feStructuredType = stNotInStructuredType) then
          feDeclarationSection := dsNotInDeclaration;
      end;
      wBegin:
        feDeclarationSection := dsNotInDeclaration;
    end
  end;
end;

procedure TTokenContext.CheckBracketLevel(const pt: TToken);
begin
  case pt.TokenType of
    ttOpenBracket: Inc(fiBracketLevel);
    ttCloseBracket: Dec(fiBracketLevel);
    ttOpenSquareBracket: Inc(fiSquareBracketLevel);
    ttCloseSquareBracket: Dec(fiSquareBracketLevel);
  end;

  if fiBracketLevel < 0 then
    fiBracketLevel := 0;
  if fiSquareBracketLevel < 0 then
    fiSquareBracketLevel := 0;
end;

procedure TTokenContext.CheckASM(const pt: TToken);
begin
  if pt.Word = wASM then
    fbASMBlock := True;

  if fbASMBlock and (pt.Word = wEnd) then
    fbASMBlock := False;
end;


procedure TTokenContext.CheckClassDefinitionSection(const pt: TToken);
begin
  // look for the word 'class' here for class functions
  if feDeclarationSection = dsNotIndeclaration then
  begin
    if (pt.Word = wClass) and (feFileSection = fsImplementation) and
      (feDeclarationSection = dsNotInDeclaration) then
      fbClassFunction := True;
  end;

  // in a class body already, the word 'class' means a class fn declaration
  if (feClassDefinitionSection in AccessSpecifierSections) then
  begin
    if (pt.Word = wClass) then
      fbClassFunction := True;
  end;

  { the rest only applies to structured type definitions }
  if (feDeclarationSection = dsType) then
  begin
    if (feStructuredType = stNotInStructuredType) then
    begin
      { beginning of class definition ? }

      {this could be a class def or a forward.
       The difference is that a forward ends with a semicolon, e.g.
       TMark = class;
       so can only start class if there is no Semicolon Next

       could be a class ref type (e.g. type TClassRef = class of TObject;)
       so look for the of next
       }

      if fbRHSEquals and not (TokenNext(wOf)) then
      begin
        if (pt.Word = wClass) then
        begin
          feStructuredType := stClass;
          feClassDefinitionSection := cdHeader;
        end;

        if (pt.Word = wRecord) then
        begin
          feStructuredType := stRecord;
          fiRecordCases    := 0;
        end;

        if (pt.Word in InterfaceWords) then
        begin
          feStructuredType := stInterface;
          feClassDefinitionSection := cdHeader;
        end;
      end;
    end
    { not in structured type }
    else if (feStructuredType in ObjectTypes) then
    begin
      if (feClassDefinitionSection = cdHeader) and (pt.TokenType = ttOpenBracket) then
        feClassDefinitionSection := cdHeritage;

      { in class/record/interface definition already
        This code here allows records & interfaces with access level specifiers
        But hey, if the input file compiles then why should I sweat it?
      }

      case pt.Word of
        wPrivate:
          feClassDefinitionSection := cdPrivate;
        wProtected:
          feClassDefinitionSection := cdProtected;
        wPublic:
          feClassDefinitionSection := cdPublic;
        wPublished:
          feClassDefinitionSection := cdPublished;
        wEnd:
        { no other use of 'end' in class/interface def? }
        begin
          feClassDefinitionSection := cdNotInClassDefinition;
          feStructuredType         := stNotInStructuredType;
          fbClassFunction          := False;
          // this is done elsewhere in generic processing of 'end' fiNestingLevel := 0;
        end;
      end;
    end
    else
    { in a strucured type, but not an object type }
    begin
      if pt.Word = wEnd then
      begin
        { in records, any variant parts (case keyword) increment the nesting level
          reset all that at the end of the record }
        fiNestingLevel   := fiNestingLevel - fiRecordCases;
        fiRecordCases    := 0;
        feStructuredType := stNotInStructuredType;
      end;
    end;
  end;
end;

procedure TTokenContext.CheckForward(const pt: TToken);
begin
  if (pt.Word in ProcedureWords) and (pt.ProcedureHasBody) then
  begin
    if FindProcedureForward then
      fbForward := True
  end;
end;

{-------------------------------------------------------------------------------
  the divsor - equip the token with the current state }


procedure TTokenContext.SetContextOnToken(const pt: TToken);
begin
  pt.FileSection      := feFileSection;
  pt.ProcedureSection := feProcedureSection;
  pt.DeclarationSection := feDeclarationSection;
  pt.ClassDefinitionSection := feClassDefinitionSection;
  pt.StructuredType   := feStructuredType;

  pt.NestingLevel       := fiNestingLevel;
  pt.ProcedureNestingLevel := fiProcedureNestingLevel;
  pt.BracketLevel       := fiBracketLevel;
  pt.SquareBracketLevel := fiSquareBracketLevel;

  pt.InUsesClause := bInUsesClause;
  pt.RHSEquals    := fbRHSEquals;
  pt.RHSAssign    := fbRHSAssign;
  pt.RHSColon     := fbRHSColon;
  pt.RHSIn        := fbRHSIn;
  pt.RHSofExpressionWord := feRHSofExpressionWord;
  pt.ASMBlock     := fbASMBlock;
  pt.RHSOperand   := fbRHSOperand;
  pt.HasForward   := fbForward;

  pt.InPropertyDefinition := fbInPropertyDefinition;
  pt.InExports            := fbInExports;
  pt.DontFormat           := fbDontFormat;
  pt.ClassFunction        := fbClassFunction;

  pt.HasContext := True;
  pt.TokenIndex := fiCount;
end;

{-------------------------------------------------------------------------------
  checks that happen after the token has context }


procedure TTokenContext.CheckProcedureSectionAfterToken;
begin
  { end a procedure  - do this after token? }
  if (fiNestingLevel <= 1) and (pt.Word = wEnd) and
    (feProcedureSection = psProcedureBody) then
  begin
    dec(fiProcedureNestingLevel);
    if (fiProcedureNestingLevel < 0) then
      fiProcedureNestingLevel := 0;

    if fiProcedureNestingLevel = 0 then
    begin
      feProcedureSection := psNotInProcedure;
      fbClassFunction    := False;
    end
    else
      { exit nested procedure body, back to declarations }
      feProcedureSection := psProcedureDeclarations
  end;
end;

{ used to distinguish unary operators from binary operators
 as they have different spacing
 it is simple in the exp.  'a := 1 - 2;' the '-' has a 1 on the left,
 ie it is on the RHS of an operand
 as opposed to a :=  -2;

 Can distingish between unary and binary, as a binary operator
 is always on the RHS of the first operand
 }

procedure TTokenContext.CheckOperand(const pt: TToken);
begin
  case pt.TokenType of
    ttSemiColon, ttAssign, ttReservedWord,
    ttOpenBracket, ttOpenSquarebracket, ttOperator:
      fbRHSOperand := False;
    ttNumber, ttLiteralString, ttWord,
    ttCloseBracket, ttCloseSquareBracket:
      fbRHSOperand := True;
  end;
end;

procedure TTokenContext.CheckNestingLevel(const pt: TToken);
begin
  { nesting level incremented from here onwards, not including the begin token }
  if (pt.Word in IndentWords) then
    inc(fiNestingLevel);

  { case .. of indent start with the of }
  if (pt.Word = wOf) and (pt.RHSOfExpressionWord = wCase) then
    inc(fiNestingLevel);

  { nesting level decremented from here onwards, not including the end token }
  if pt.Word in OutdentWords then
  begin
    Assert(fiNestingLevel >= 0, 'Nesting level below zero in ' + OriginalFileName);
    dec(fiNestingLevel);
    { just in case there is an asign statement with no semicolon,
    eg  begin a:= 12 end }

    fbRHSAssign := False;
  end;

  {  'class' is only sometimes an indent word -
     indent when starting a class def but not on 'class' for a proc
     interface and dispinterface }
  if (feDeclarationSection = dsType) and (feStructuredType <> stNotInStructuredType) and
    (pt.Word in ObjectTypeWords) and (not pt.ClassFunction) and fbRHSEquals then
    fiNestingLevel := 1;

{
  THis test produces many errors here
  but the equivalent one produces none in the indenter???
  if fiNestingLevel < 0 then
  begin
    Log.LogError('Nesting level is ' + IntToStr(fiNestingLevel) +
      ' Resetting to zero in ' + FilePlace(pt) +  ' near ' +
        pt.Describe);
    fiNestingLevel := 0;
  end;
}
end;

procedure TTokenContext.CheckEndRHS(const pt: TToken);
var
  ptNext: TToken;
begin
  { tokens *after* the = sign have the flag set }
  if pt.Word = wEquals then
    fbRHSEquals := True;

  if pt.TokenType = ttAssign then
    fbRHSAssign := True;

  if pt.TokenType = ttColon then
    fbRHSColon := True;

  if (pt.Word = wProperty) and (pt.StructuredType in ObjectTypes) then
    fbInPropertyDefinition := True;

  if (pt.Word = wExports) then
    fbInExports := True;

  if pt.TokenType = ttSemiColon then
  begin
    { const record defs don't end at the first semicolon. Semicolons seperate fields }
    if (not pt.IsSeparatorSemiColon) and (not pt.IsParamSemicolon) then
      fbRHSEquals := False;
      
    fbRHSAssign := False;
    fbRHSColon  := False;
    feRHSofExpressionWord := wUnknown;
    { property def doesn't end if there is the word default after the semicolon :(
     used to declare a default array property }

    if fbInPropertyDefinition then
    begin
      ptNext := FirstSolidToken;
      { still in the property def *only* if 'default' is next }
      fbInPropertyDefinition := (ptNext.Word = wDefault);
    end;
    fbInExports := False;
  end;
end;

procedure TTokenContext.CheckClassHeader(const pt: TToken);
  procedure EndOfClassHeritage;
  begin
    // end of heritage
    fbRHSEquals := False;

    if SemiColonNext then
    begin
      feClassDefinitionSection := cdNotInClassDefinition;
      feStructuredType         := stNotInStructuredType;
      fiNestingLevel           := 0; // as end;
    end
    else
      feClassDefinitionSection := cdPublic; // public by default AFAIK
  end;
begin
  { End the class/interface heritage and start the body
    Signaled either by the close bracket or the lack of open bracket, e.g.
      type
        TFred = class
          function Fred11;
        end;

    Also end the RHSEquals, as there is no semicolon to end it otherwise e.g.
    type
      TMyClass = class (TObject)
        procedure Foo (pi: integer); // this line should not be considered RHSEqals

      TMyOtherClass = class
        procedure Foo (pi: integer); // this line should not be considered RHSEqals

    and it must be false before the procs start

    also class with no body -
      type TFoo = class; is a forward, but type TBar = class(TObject); is a subclass
    but in both cases the def. ends with the semicolon
    }

  if fbRHSEquals and (feClassDefinitionSection in ClassStartSections) then
  begin
    if (feClassDefinitionSection = cdHeader) and (pt.Word in ObjectTypeWords)
      and not TokenNext(ttOpenBracket) then
      EndOfClassHeritage;

    if (feClassDefinitionSection = cdHeritage) and (pt.TokenType = ttCloseBracket) then
    begin
      EndOfClassHeritage;
    end;
  end;
end;

procedure TTokenContext.CheckNoFormat(const pt: TToken);
begin
  { don't format specifers
   I am using the same signals as Egbert van Nes's program }
  if (pt.TokenType = ttComment) then
  begin
    if pt.SourceCode = NOFORMAT_ON then
      fbDontFormat := True;
    if pt.SourceCode = NOFORMAT_OFF then
      fbDontFormat := False;
  end;
end;

procedure TTokenContext.StartProcNestingLevel(const pt: TToken);
begin
  if (pt.Word in ProcedureWords) and (pt.ProcedureHasBody) then
    inc(fiProcedureNestingLevel);
end;

procedure TTokenContext.CheckRHSExpression(const pt: TToken);
begin
  if (feProcedureSection = psProcedureBody) then
  begin
    { end a Word pair }
    if (pt.Word = wThen) and (pt.RHSOfExpressionWord = wIf) then
    begin
      feRHSOfExpressionWord := wUnknown;
      // if..then frequently involves a =
      fbRHSEquals           := False;
    end;
    if (pt.TokenType = ttSemiColon) and (pt.RHSOfExpressionWord = wUntil) then
      feRHSOfExpressionWord := wUnknown;
    if (pt.Word = wDo) and (pt.RHSOfExpressionWord = wWhile) then
      feRHSOfExpressionWord := wUnknown;
    if (pt.Word = wOf) and (pt.RHSOfExpressionWord = wCase) then
      feRHSOfExpressionWord := wUnknown;
    if (pt.Word = wDo) and (pt.RHSOfExpressionWord = wFor) then
    begin
      feRHSOfExpressionWord := wUnknown;
      { for loop has 'for liVar := liStart to liEnd do ... '
       so we are RHSAssign here, stop that at the do }

      fbRHSAssign := False;
    end;

    { start a word pair }
    if (pt.Word in PairedWords) then
      feRHSOfExpressionWord := pt.Word;
  end;

  { you get case in records }
  if (feDeclarationSection = dsType) and (feStructuredType = stRecord) then
  begin
    if pt.Word = wCase then
    begin
      feRHSOfExpressionWord := wCase;
      inc(fiRecordCases);
    end;
  end;
end;

procedure TTokenContext.CheckImplicitBlocks(const pt: TToken);
begin
  { intialization and finalization sections are just code
  ie they are implicitly procs
    enter an implicit block when entering them
  }

  if (feFileSection in CodeSections) and (pt.Word in SectionWords) and
    (pt.ProcedureSection = psNotInProcedure) then
    SetImplicitBlock;

  { program and library sections have a begin..end. at the bottom of the unit
    which is the top of the executable progam
    start indenting *after* the begin
  }

  if (feFileSection in ProgramDefSections) and (pt.Word = wBegin) and
    not (fbImplicitBlock) and (pt.ProcedureSection = psNotInProcedure) then
    SetImplicitBlock;
end;

procedure TTokenContext.CheckUsesClauseAfter(const pt: TToken);
begin
  { uses clause termination stuff }
  if pt.TokenType = ttSemiColon then
  begin
    bInUsesClause := False;
    fbRHSIn       := False;
  end;
  if (pt.TokenType = ttComma) and bInUsesClause then
    fbRHSIn := False;
end;

end.