//====================================================
//
//  Wizard component by ztn
//
//  ztn@planetquake.com
//  http://www.planetquake.com/ztn
//
//----------------------------------------------------
//  Version history:
//    9/ 6/2000 - 1st public release
//    9/26/2000 - declared some methods as virtual
//                added OnUpdateHeaders event
//====================================================

unit wizards;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls;

resourcestring
  sWizardButtonNextCaption   = '&Next >';
  sWizardButtonBackCaption   = '< &Back';
  sWizardButtonFinishCaption = '&Finish';

type
  TWizardDirection = (wdForward, wdBackward);

  TWizardButtonClickEvent = procedure (Sender: TObject; CurrentPage: Integer;
    var AllowMove: Boolean) of object;
  TEnteringPageEvent = procedure (Sender: TObject; CurrentPage: Integer;
    Direction: TWizardDirection) of object;
  TUpdateHeadersEvent = procedure (Sender: TObject; HeaderText, InfoText: string;
    CurrentPage: Integer) of object;

  TWizard = class(TNotebook)
  private
    FButtonNext: TButton;
    FButtonBack: TButton;

    FStaticTextHeader: TStaticText;
    FStaticTextInfo: TStaticText;

    FHeaderCaption: TStrings;
    FInfoCaption: TStrings;
    
    FFinishing: Boolean;
    FHideButtons: Boolean;
    FPageIndex: Integer;

    FOnButtonBackClick: TWizardButtonClickEvent;
    FOnButtonNextClick: TWizardButtonClickEvent;
    FOnButtonFinishClick: TNotifyEvent;
    FOnEnteringPage: TEnteringPageEvent;
    FOnUpdateHeaders: TUpdateHeadersEvent;

    procedure AdjustButtonCaptions;
    function GetBackEnabled: Boolean;
    function GetFinishing: Boolean;
    function GetNextEnabled: Boolean;
    function GetNextPage: Integer;
    function GetPrevPage: Integer;
    function GetPageIndex: Integer;
    procedure SetHeaderCaption(Value: TStrings);
    procedure SetHideButtons(Value: Boolean);
    procedure SetInfoCaption(Value: TStrings);
    procedure SetStaticTextHeader(const Value: TStaticText);
    procedure SetStaticTextInfo(const Value: TStaticText);
    procedure SetButtonBack(const Value: TButton);
    procedure SetButtonNext(const Value: TButton);
    procedure SetNextEnabled(const Value: Boolean);
    procedure SetBackEnabled(const Value: Boolean);
    procedure SetPageIndex(const Value: Integer);
  protected
    procedure ButtonBackClick(Sender: TObject); virtual;
    procedure ButtonNextClick(Sender: TObject); virtual;
    procedure DisableButton(const Button: TButton);
    procedure EnableButton(const Button: TButton);
    procedure Loaded; override;
    procedure UpdateHeaders; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IsSkipPage(PageNum: Integer): Boolean; virtual;
    procedure SetSkipPage(PageNum: Integer; Skip: Boolean); virtual;

    property Finishing: Boolean read GetFinishing;
    property BackEnabled: Boolean read GetBackEnabled write SetBackEnabled;
    property NextEnabled: Boolean read GetNextEnabled write SetNextEnabled;
  published
    property ButtonNext: TButton read FButtonNext write SetButtonNext;
    property ButtonBack: TButton read FButtonBack write SetButtonBack;
    property HeaderCaption: TStrings read FHeaderCaption write SetHeaderCaption;
    property HideButtons: Boolean read FHideButtons write SetHideButtons default False;
    property InfoCaption: TStrings read FInfoCaption write SetInfoCaption;
    property PageIndex: Integer read GetPageIndex write SetPageIndex;
    property StaticTextHeader: TStaticText read FStaticTextHeader write SetStaticTextHeader;
    property StaticTextInfo: TStaticText read FStaticTextInfo write SetStaticTextInfo;

    property OnButtonBackClick: TWizardButtonClickEvent read FOnButtonBackClick write FOnButtonBackClick;
    property OnButtonNextClick: TWizardButtonClickEvent read FOnButtonNextClick write FOnButtonNextClick;
    property OnButtonFinishClick: TNotifyEvent read FOnButtonFinishClick write FOnButtonFinishClick;
    property OnEnteringPage: TEnteringPageEvent read FOnEnteringPage write FOnEnteringPage;
    property OnUpdateHeaders: TUpdateHeadersEvent read FOnUpdateHeaders write FOnUpdateHeaders;
  end;

procedure Register;

implementation

{ Global }

procedure Register;
begin
  RegisterComponents('Custom', [TWizard]);
end;

{ TWizard }

constructor TWizard.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHeaderCaption := TStringList.Create;
  FInfoCaption := TStringList.Create;
end;

destructor TWizard.Destroy;
begin
  FHeaderCaption.Free;
  FInfoCaption.Free;
  inherited Destroy;
end;

procedure TWizard.AdjustButtonCaptions;
begin
  if not (csDesigning in ComponentState) and (FButtonNext <> nil) then
    with FButtonNext do
      if Finishing then Caption := sWizardButtonFinishCaption
      else Caption := sWizardButtonNextCaption;
end;

function TWizard.IsSkipPage(PageNum: Integer): Boolean;
begin
  if (Pages.Count > 0) and (PageNum < Pages.Count) and
     (Pages.Objects[PageNum] <> nil) then
    Result := (Pages.Objects[PageNum] as TCustomControl).HelpContext < 0
  else
    Result := False;
end;

procedure TWizard.SetSkipPage(PageNum: Integer; Skip: Boolean);
begin
// Skipping pages is done by setting the HelpContext value
// on the indicated page to a negative number, or to the lowest
// possible value of HelpContext if HelpContext was zero.
// You can't skip the first and last page
  if (Pages.Count > 2) and (PageNum > 0) and (PageNum < Pages.Count - 1) and
     (Pages.Objects[PageNum] <> nil) then
    with Pages.Objects[PageNum] as TCustomControl do
      if Skip then
      begin
        if HelpContext = 0 then HelpContext := Low(HelpContext)
        else if HelpContext > 0 then HelpContext := -HelpContext;
      end else
        if HelpContext = Low(HelpContext) then HelpContext := 0
        else if HelpContext < 0 then HelpContext := -HelpContext;
end;

procedure TWizard.SetButtonNext(const Value: TButton);
begin
  FButtonNext := Value;
  if not (csDesigning in ComponentState) and (FButtonNext <> nil) then
  begin
    FButtonNext.OnClick := ButtonNextClick;
    FFinishing := Pages.Count = 1;
    if FFinishing then
      FButtonNext.Caption := sWizardButtonFinishCaption
    else
      FButtonNext.Caption := sWizardButtonNextCaption;
  end
end;

procedure TWizard.SetButtonBack(const Value: TButton);
begin
  FButtonBack := Value;
  if not (csDesigning in ComponentState) and (FButtonBack <> nil) then
  begin
    FButtonBack.OnClick := ButtonBackClick;
    FButtonBack.Caption := sWizardButtonBackCaption;
    DisableButton(FButtonBack);
  end;
end;

procedure TWizard.SetHeaderCaption(Value: TStrings);
begin
  FHeaderCaption.Assign(Value);
  UpdateHeaders;
end;

procedure TWizard.SetHideButtons(Value: Boolean);
begin
  FHideButtons := Value;
  if csDesigning in ComponentState then Exit;
  if FButtonNext <> nil then
    FButtonNext.Visible := not FHideButtons or NextEnabled;
  if FButtonBack <> nil then
    FButtonBack.Visible := not FHideButtons or BackEnabled;
end;

procedure TWizard.SetInfoCaption(Value: TStrings);
begin
  FInfoCaption.Assign(Value);
  UpdateHeaders;
end;

procedure TWizard.SetStaticTextHeader(const Value: TStaticText);
begin
  FStaticTextHeader := Value;
  UpdateHeaders;
end;

procedure TWizard.SetStaticTextInfo(const Value: TStaticText);
begin
  FStaticTextInfo := Value;
  UpdateHeaders;
end;

function TWizard.GetNextPage: Integer;
begin
  // Find the next page which isn't marked as skip page
  Result := FPageIndex + 1;
  if Pages.Count > 2 then
    while (Result < Pages.Count - 1) and IsSkipPage(Result) do
      Inc(Result);
  if Result > Pages.Count - 1 then
    Result := Pages.Count - 1;
end;

function TWizard.GetPrevPage: Integer;
begin
  // Find the previous page which isn't marked as skip page
  Result := FPageIndex - 1;
  if Pages.Count > 2 then
    while (Result > 0) and IsSkipPage(Result) do
      Dec(Result);
  if Result < 0 then
    Result := 0;
end;

function TWizard.GetPageIndex: Integer;
begin
  FPageIndex := Pages.IndexOf(ActivePage);
  Result := FPageIndex;
end;

function TWizard.GetBackEnabled: Boolean;
begin
  if FButtonBack <> nil then Result := FButtonBack.Enabled
  else Result := False;
end;

function TWizard.GetFinishing: Boolean;
begin
  Result := (Pages.Count <= 1) or (FPageIndex = Pages.Count - 1);
end;

function TWizard.GetNextEnabled: Boolean;
begin
  if FButtonNext <> nil then Result := FButtonNext.Enabled
  else Result := False;
end;

procedure TWizard.SetNextEnabled(const Value: Boolean);
begin
  // Enable/disable the 'Next' button
  if FButtonNext <> nil then
    if Value then EnableButton(FButtonNext)
    else DisableButton(FButtonNext)
end;

procedure TWizard.SetBackEnabled(const Value: Boolean);
begin
  // Enable/disable the 'Back' button
  if FButtonBack <> nil then
    if Value then EnableButton(FButtonNext)
    else DisableButton(FButtonNext)
end;

procedure TWizard.SetPageIndex(const Value: Integer);
begin
  // This method overrides our descendant class' 'PageIndex := Value'
  // statement. That's because we want to call the UpdateHeaders to make
  // sure the headers reflect the page changes, too
  if (Value <> FPageIndex) and (Value >= 0) and (Value < Pages.Count) then
  begin
    FPageIndex := Value;
    ActivePage := Pages[FPageIndex];
    UpdateHeaders;
  end;
end;

procedure TWizard.ButtonBackClick(Sender: TObject);
var
  PrevIndex: Integer;
  AllowMove,
  OnFirstPage: Boolean;
begin
  // This method gets assinged to your BackButton's
  // OnClick event when assinging your buttons to TWizard.
  // That's also why you shouldn't declare your own OnClick
  // events to your Back/Next buttons -- they'll get overrided.
  // In such cases use TWizard's OnButtonBackClick/OnButtonNextClick
  // events instead
  if Assigned(FOnButtonBackClick) then
  begin
    AllowMove := True;
    FOnButtonBackClick(Sender, PageIndex, AllowMove);
    if not AllowMove then Exit;
  end;
  OnFirstPage := (PageIndex = 0) or (Pages.Count = 1);
  if not OnFirstPage then
  begin
    PrevIndex := FPageIndex;
    PageIndex := GetPrevPage;
    if FPageIndex = PrevIndex then Exit;
    EnableButton(FButtonNext);
    AdjustButtonCaptions;
    // Check to see if we are on first page after the move
    if (FPageIndex = 0) and (Pages.Count > 1) then
      DisableButton(FButtonBack);
    if Assigned(FOnEnteringPage) then
      FOnEnteringPage(Sender, FPageIndex, wdBackward);
  end else
    DisableButton(FButtonBack);
end;

procedure TWizard.ButtonNextClick(Sender: TObject);
var
  PrevIndex: Integer;
  AllowMove: Boolean;
begin
  // This method gets assinged to your NextButton's
  // OnClick event when assinging your buttons to TWizard.
  // That's also why you shouldn't declare your own OnClick
  // events to your Back/Next buttons -- they'll get overrided.
  // In such cases use TWizard's OnButtonBackClick/OnButtonNextClick
  // events instead
  if not Finishing then
  begin
    if Assigned(FOnButtonNextClick) then
    begin
      AllowMove := True;
      FOnButtonNextClick(Sender, PageIndex, AllowMove);
      if not AllowMove then Exit;
    end;
    PrevIndex := PageIndex;
    PageIndex := GetNextPage;
    if FPageIndex = PrevIndex then Exit;
    EnableButton(FButtonBack);
    AdjustButtonCaptions;
    if Assigned(FOnEnteringPage) then
      FOnEnteringPage(Sender, FPageIndex, wdForward);
  end else
    if Assigned(FOnButtonFinishClick) then
      FOnButtonFinishClick(Self);
end;

procedure TWizard.DisableButton(const Button: TButton);
begin
  if Button <> nil then
    with Button do
    begin
      Enabled := False;
      if FHideButtons then Visible := False;
    end;
end;

procedure TWizard.EnableButton(const Button: TButton);
begin
  if Button <> nil then
  begin
    Button.Visible := True;
    Button.Enabled := True;
  end;
end;

procedure TWizard.Loaded;
begin
  inherited;
  if not (csDesigning in ComponentState) then
    PageIndex := 0;
end;

procedure TWizard.UpdateHeaders;

  procedure DoSetHeader(ST: TStaticText; List: TStrings; var Txt: string);
  begin
    Txt := '';
    with List do
      if PageIndex < Count then Txt := Strings[FPageIndex];
    if ST <> nil then
      ST.Caption := Txt;
  end;

var
  HeaderText,
  InfoText: string;
begin
  if not (csDesigning in ComponentState) then
  begin
    DoSetHeader(FStaticTextHeader, FHeaderCaption, HeaderText);
    DoSetHeader(FStaticTextInfo, FInfoCaption, InfoText);
    if Assigned(FOnUpdateHeaders) then
      FOnUpdateHeaders(Self, HeaderText, InfoText, FPageIndex);
  end;
end;

end.
